import numpy as np
import logging
logger = logging.getLogger('global')
import pandas as pd




class EPFLReader(object):
    def __init__(self, filename):
        '''
        Parse EPFL Challenge dataset
        :param filename: Path to file
        '''
        self._filename = filename
        self._values = None
        self._points = None
        logger.debug("Starting decode for {}".format(filename))
        self._read_ascii()
        logger.debug("Complete")


    def _read_ascii(self):
        data = pd.read_csv(self._filename)
        pts = data[['xnano', 'ynano', 'znano']]
        assert(pts.shape[1] == 3)
        vls = data[['frame', 'Ground-truth', 'intensity ']]
        self._points=pts.values.copy()
        self._values=vls.values.copy()
        self._columns = ['frame', 'Ground-truth', 'intensity ']


    @property
    def points(self):
        return self._points

    @property
    def values(self):
        return self._values

    @property
    def value_names(self):
        return self._columns

    def points_generator(self):
        cnt = 0
        l = len(self.points)
        while cnt < l:
            yield self._points[cnt]

    def values_generator(self):
        cnt = 0
        l = len(self.values)
        while cnt < l:
            yield self._values[cnt]



import vtk
import numpy as np

class TemporalVtuWriter(object):
    def __init__(self, filename, points, values, incremental=False, limit=0, collate_frames=1):
        '''
        Writes filename<number>.vtu of points, values sliced along frames.
        The slicenumber corresponds with the last fields of the k-tuple in values.
        '''
        self._filename = filename
        self._points = points
        self._values = values
        self._incremental = incremental
        self._limit = limit
        self._skipframes = 1 if collate_frames <= 1 else collate_frames
        self._writeAll()

    def _writeAll(self):
        fname = self._filename
        # find out if doing binary split 40k times is more efficient
        lastframe = 1
        lastindex = 0
        for index, (point, value) in enumerate(zip(self._points, self._values)):
            fnumber = int(value[-1])
            if fnumber > lastframe + self._skipframes - 1:
                # print("New framenumber from {} to {} with index going from {} to {}".format(lastframe, fnumber, lastindex, index))
                l = 0 if self._incremental else lastindex
                p, v = self._points[l:index, :], self._values[l:index, :]
                sfname = "{}{}".format(fname, fnumber-1)
                VtuWriter(sfname, p, v)
                lastframe, lastindex = fnumber, index
                if self._limit != 0 and fnumber > self._limit:
                    return


class VtuWriter(object):
    def __init__(self, filename, points, values):
        '''
        Appends .vtu to filename
        '''
        self._points= vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(points, values)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, points, values):
        poly = vtk.vtkPolyVertex()
        poly.GetPointIds().SetNumberOfIds(points.shape[0])
        for i, (point, value) in enumerate(zip(points, values)):
            poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
            self._values.InsertNextValue(values[i][-1])
        self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())

    def _write(self, filename):
        writer = vtk.vtkXMLUnstructuredGridWriter()
        writer.SetFileName(filename)
        writer.SetInputData(self._grid)
        writer.Write()


class GSDWriter(VtuWriter):
    def __init__(self, filename, points, values):
        '''
        Appends .vtu to filename
        '''
        self._points= vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(points, values)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, points, values):
        poly = vtk.vtkPolyVertex()
        poly.GetPointIds().SetNumberOfIds(points.shape[0])
        for i, (point, value) in enumerate(zip(points, values)):
            poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
            self._values.InsertNextValue(values[i][-3]) # fix arbitrary data
        self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())


class KChannelVtuWriter(VtuWriter):
    def __init__(self, filename, points, values, channels):
        self._channels = channels
        self._points = vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(points, values)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, pointset, valueset): # values is set
        for index, (points, _, values) in enumerate(zip(pointset, self._channels, valueset)):
            poly = vtk.vtkPolyVertex()
            poly.GetPointIds().SetNumberOfIds(points.shape[0])
            for i, (point, value) in enumerate(zip(points, values)):
                poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
                self._values.InsertNextValue(index)
            self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())


class NNWriter(VtuWriter):
    def __init__(self, filename, points, values, channels, edges):
        '''
        Writes a k-way 1-nearest neighbour graph between points using edges.
        '''
        self._channels = channels
        self._points = vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        self._edges = edges # Dict-> {offset_src, offset_trg} -> [(src_index, trg_index)]
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(points, values)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, pointset, valueset): # values is set
        # Points
        for index, (points, _, values) in enumerate(zip(pointset, self._channels, valueset)):
            poly = vtk.vtkPolyVertex()
            poly.GetPointIds().SetNumberOfIds(points.shape[0])
            for i, (point, value) in enumerate(zip(points, values)):
                poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
                self._values.InsertNextValue(index)
            self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())
        # Edges
        for offset, edges in self._edges.items():
            sourceoffset, targetoffset = offset
            for source, target in edges:
                self._grid.InsertNextCell(vtk.VTK_LINE, 2, [sourceoffset+source, targetoffset+target])


class GraphWriter(VtuWriter):
    def __init__(self, filename, midpoints, values, tetras):
        '''
        Writes a k-way 1-nearest neighbour graph between points using edges.
        '''
        # self._channels=channels
        self._points= vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        # self._edges = edges # Dict-> {offset_src, offset_trg} -> [(src_index, trg_index)]
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(midpoints, values, tetras)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, points, values, tetras):
        # Points
        poly = vtk.vtkPolyVertex()
        poly.GetPointIds().SetNumberOfIds(points.shape[0])
        for i, (point, value) in enumerate(zip(points, values)):
            poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
            self._values.InsertNextValue(value)
        self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())
        # Tetras
        # print('Adding {} tetras'.format(len(tetras)))
        for tetra in tetras:
            self._grid.InsertNextCell(vtk.VTK_TETRA, 4, tetra)


class AlphaGraphWriter(VtuWriter):
    def __init__(self, filename, midpoints, values, tetras):
        '''
        Writes a k-way 1-nearest neighbour graph between points using edges.
        '''
        # self._channels=channels
        self._points= vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        # self._edges = edges # Dict-> {offset_src, offset_trg} -> [(src_index, trg_index)]
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(midpoints, values, tetras)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, points, values, tetras):
        # Points
        poly = vtk.vtkPolyVertex()
        poly.GetPointIds().SetNumberOfIds(points.shape[0])
        for i, (point, value) in enumerate(zip(points, values)):
            poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
            self._values.InsertNextValue(value)
        self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())
        # Tetras
        # print('Adding {} tetras'.format(len(tetras)))
        for tetra in tetras:
            self._grid.InsertNextCell(vtk.VTK_TETRA, 4, tetra)


class MatWriter(VtuWriter):
    def __init__(self, filename, mat, flip = True):
        '''
        Write a square matrix to VTK.
        Flip results in [0,0] to [n,n] being mapped as top left, lower right.
        '''
        assert(mat is not None)
        self._points = vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints(mat)
        self._write("{}.vtu".format(filename))

    def _loadPoints(self, mat):
        n = mat.shape[0]
        nsq = n**2
        points = np.empty((nsq, 3))
        values = np.empty((nsq, 1))
        for k in range(nsq):
            i, j = divmod(k, n)
            points[k] = [n-i,j,0]
            values[k] = mat[i,j] if (i,j) in mat else 0

        poly = vtk.vtkPolyVertex()
        poly.GetPointIds().SetNumberOfIds(points.shape[0])
        for i, (point, value) in enumerate(zip(points, values)):
            poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
            self._values.InsertNextValue(value)
        self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())


class MidPointWriter(VtuWriter):
    def __init__(self, filename, midpoints):
        '''
        Writes a point cloud of k midpoints
        '''
        self._k = len(midpoints)
        self._points = vtk.vtkPoints()
        self._grid = vtk.vtkUnstructuredGrid()
        self._values = vtk.vtkDoubleArray()
        self._midpoints = midpoints
        self._values.SetName('point_values_array')
        self._grid.SetPoints(self._points)
        self._grid.GetPointData().SetScalars(self._values)
        self._loadPoints()
        self._write("{}.vtu".format(filename))

    def _loadPoints(self):
        for (src_offset, trg_offset),midpoints in self._midpoints.items():
            p, v = zip(*midpoints)
            points = np.empty((len(p), len(p[0])))
            for index, coord in enumerate(p):
                points[index, :] = coord
            values = np.empty((len(v), len(v[0])))
            for index, value in enumerate(v):
                values[index, :] = value
            poly = vtk.vtkPolyVertex()
            poly.GetPointIds().SetNumberOfIds(points.shape[0])
            for i, (point, value) in enumerate(zip(points, values)):
                poly.GetPointIds().SetId(i, self._points.InsertNextPoint(point))
                self._values.InsertNextValue(value[-1]) # Use distance as value
            self._grid.InsertNextCell(poly.GetCellType(), poly.GetPointIds())



import numpy as np
import struct
import logging
logger = logging.getLogger('global')


def type_to_unpack(name):
    d = {'UINT32':'I', 'float':'f'}
    return d[name]

def bytelength(name):
    d = {'I':4, 'f': 4}
    return d[name]

def desctotypes(filename):
    with open(filename, 'r') as f:
        return [(word[:word.index('(')].strip(), type_to_unpack(word[word.index('(')+1:word.index(')')])) for word in f.readline().split(',')]


def mask_z(points, values, threshold):
    '''
    :param points: np array 3xN
    :param values: np array 6xN
    :param threshold: Z values, all rows where abs(points[:,2]) exceeds treshold are omitted
    :return: truncated points, values
    '''
    n = len(points)
    mask = abs(points[:,2]) < threshold
    trimmedp, trimmedv = points[mask].copy(), values[mask].copy()
    diff = n - len(trimmedp)
    pct = (diff / n) * 100
    # logger.info('Removed {} points, {} % of data'.format(diff , pct))
    return trimmedp, trimmedv, pct


def gsd_preprocess(points, values):
    p, v, pct = mask_z(points, values, 1e15) # GSD has corrupt Z values (~-500 to 500 with outliers 1e36
    return p, v, pct

def filter_photoncount_outliers(points, values, k_sd=3):
    '''
    Photoncount filter for GSD
    :param points: N x 3 numpy array
    :param values: N x 6 numpy array where col [3] is photoncount
    :param k_sd: all points < mean + k_sd * sd are filtered out
    :return: p', v'
    '''
    pc = values[:,3]
    pm, ps = np.mean(pc), np.std(pc)
    threshold = pm + k_sd*ps
    logger.info('Treshold {}'.format(threshold))
    m = values[:,3] <= threshold
    pk, vk = points[m].copy(), values[m].copy()
    logger.info('Removed {} % points'.format(100 - len(pk)/len(points) * 100))
    return pk, vk

def filter_z_plane(points, values, z, Z):
    '''
    :param points: N x 3 np array
    :param values: N x 6 np array
    :param z: min z value tolerated
    :param Z: max z value tolerated
    :return: A filtered copy of points, values
    '''
    assert(points.shape[1] == 3)
    logger.info('Min {} Max {}'.format(z, Zb))
    m = (points[:,2] > z) & (points[:,2] <= Z)
    return points[m].copy(), values[m].copy()


def slice_roi(points, values, roimin, roimax):
    mask = None
    # True if value in accepted range
    for i, (m, M) in enumerate(zip(roimin, roimax)):
        mm = np.logical_and(points[:,i] > m, points[:,i] < M)
        if mask is not None:
            mask = np.logical_and(mask, mm)
        else:
            mask = mm
    fp = points[mask].copy()
    fv = values[mask].copy()
    N = len(points)
    d = len(fp)
    logger.info('Removed {} out of {} points'.format(N-d, N))
    return fp, fv


def split_gsd_data(fpoints, fvalues):
    x, y, z = np.split(fpoints, 3, 1)
    stack_id, frame_id, eventid, pcount, sigma_x, sigma_y = np.split(fvalues, 6, 1)
    return (x, y, z), (stack_id, frame_id, eventid, pcount, sigma_x, sigma_y)


def dilutions(tree, imnr):
    return list(tree[imnr].keys())


def antibodies(tree, imnr, dilution):
    return list(tree[imnr][dilution].keys())


def files(tree, imnr, dilution, ab):
    fs = tree[imnr][dilution][ab]
    for filename, fullname in fs.items():
        if filename.endswith('.bin'):
            return fullname


def loadcell(n, tree):
    '''
    :param n: String integer cell number
    :param tree:  Nested dict of cell number, dilution, antibody , filename
    :return: dict of data[dilution][antiobody] = (x,y,z), (stack_id, frame_id, eventid, pcount, sigma_x, sigma_y)
    '''
    data = {}
    dis = dilutions(tree, n)
    for di in dis:
        data[di] = {}
        ab_s = antibodies(tree, n, di)
        for ab in ab_s:
            bin_filename = files(tree, n, di, ab)
            r = GSDReader(bin_filename)
            raw_points, raw_values = r.points, r.values
            fpoints, fvalues = mask_z(raw_points, raw_values, 1e30)
            (x,y,z), (stack_id, frame_id, eventid, pcount, sigma_x, sigma_y) = split_gsd_data(fpoints, fvalues)
            data[di][ab]=(x,y,z), (stack_id, frame_id, eventid, pcount, sigma_x, sigma_y)
    return data


def loadcellfile(fname):
    r = GSDReader(fname)
    raw_points, raw_values = r.points, r.values
    fpoints, fvalues = mask_z(raw_points, raw_values, 1e30)
    (x, y, z), (stack_id, frame_id, eventid, pcount, sigma_x, sigma_y) = split_gsd_data(fpoints, fvalues)
    return (x, y, z), (stack_id, frame_id, eventid, pcount, sigma_x, sigma_y)


class GSDReader(object):
    '''
        Read a GSD file into memory
        Filename is the name of the binary data file, with a corresponding filename.desc file in the same location.
        This header file is parsed to get the alignment (int, float, etc).
    '''
    def __init__(self, filename, preprocess=True, binary=True):
        '''
        Parse GSD files.
        :param filename: Path to file (if binary, expects filename.desc with headers for encoding
        :param preprocess: If true, remove invalid values
        :param binary: If true, reads binary files. If False, ascii.
        '''
        self._filename = filename
        self._values = None
        self._points = None
        self._preprocess = preprocess
        logger.debug("Starting decode for {}".format(filename))
        self._binary = binary
        if self._binary:
            self._GSD_Encoding=desctotypes(filename+'.desc')
            logger.debug('Decoded types: {}'.format(self._GSD_Encoding))
            self._unpack_string = ''.join(t for _,t in self._GSD_Encoding)
            # logger.info('Unpack String: {}'.format(self._unpack_string))
            self._columns = [n for n, _ in self._GSD_Encoding]
            self._read_binary()
        else:
            self._read_ascii()
        self._post_read()

    def _post_read(self):
        if self._points is None:
            logger.error("Decoding failed, can't preprocess")
            return
        N = self._points.shape[0]
        if self._preprocess:
            # logger.debug("Pre Processing data for invalid Z coordinates")
            fp, fv, pct = gsd_preprocess(self._points, self._values)
            # logger.info('PCt {}'.format(pct))
            self._pct = pct
            self._points = fp
            self._values = fv
            logger.debug("Retained {} out {} points".format(self._points.shape[0], N))
        else:
            pass
        logger.debug('Mean X {} Mean Y {} Mean Z{}'.format(np.mean(self._points[:,0]), np.mean(self._points[:,1]), np.mean(self._points[:,2])))


    def _read_ascii(self):
        data = None
        with open(self._filename) as f:
            line = f.readline()
            if '#' in line:
                line = line[1:]
            self._GSD_Encoding = [(word[:word.index('(')].strip(), type_to_unpack(word[word.index('(')+1:word.index(')')])) for word in line.split(',')]
            self._columns = [n for n, _ in self._GSD_Encoding]
        try:
            decoded = np.loadtxt(self._filename, delimiter=',', dtype='float64')
            assert(decoded.shape[1] == 9)
            self._values = np.empty((decoded.shape[0], decoded.shape[1]-3))
            self._points = decoded[:,3:6]
            self._values[:,0:3] = decoded[:,0:3]
            self._values[:,3:] = decoded[:,6:]

        except Exception as e:
            logger.error('Failed parsing file with exception {}'.format(e))


    def _read_binary(self):
        data = None
        linelength = sum(bytelength(t) for t in self._unpack_string)

        with open(self._filename, "rb") as f:
            data = f.read()

        bytecount = len(data)
        assert(bytecount % linelength == 0)
        linecount = int(bytecount / linelength)
        # logger.debug('Have {} bytes, {} bytes per line, {} lines.'.format(bytecount, linelength, linecount))

        points = np.empty((linecount, 3))
        values = np.empty((linecount, 6))
        if data:
            for linenumber in range(linecount):
                line = data[ linenumber * linelength : (linenumber+1)*linelength]
                assert(len(line)== linelength)
                decoded = struct.unpack(self._unpack_string, line)
                x, y, z = decoded[3:6]
                points[linenumber] = x, y, z
                values[linenumber] = decoded[0:3] + decoded[6:]
            logger.info('Decoding complete of file {}'.format(self._filename))
            self._points = points
            self._values = values
        else:
            logger.error('Failed reading data.')
        # logger.debug('Mean X {} Mean Y {} Mean Z{}'.format(np.mean(self._points[:,0]), np.mean(self._points[:,1]), np.mean(self._points[:,2])))



    @property
    def points(self):
        return self._points

    @property
    def invalid_data_pct(self):
        return self._pct

    @property
    def values(self):
        return self._values

    @property
    def value_names(self):
        return self._columns

    def points_generator(self):
        cnt = 0
        l = len(self.points)
        while cnt < l:
            yield self._points[cnt]

    def values_generator(self):
        cnt = 0
        l = len(self.values)
        while cnt < l:
            yield self._values[cnt]
#
# if __name__=='__main__':
#     # r = DlpReader('../data/20170725_1_G.3dlp')
#     # p = r.points[0]
#     # print(type(p))
#     # print(p.shape)
#     # word_types = desctotypes("../../data/1000_Image1_cy3b_eventlist_July252018.bin.desc")
#     r = GSDReader("../../data/600_Image2_CY3B_eventlist_July242018.bin")
#     logger.info('Have {} points'.format(len(r.points)))


import numpy as np

class DlpReader(object):
    ''' Read a .3dlp file into memory'''
    def __init__(self, filename):
        self._filename = filename
        self._values = None
        self._points = None
        self._columns = ['std_x', 'std_y', 'std_z', 'amplitude', 'framenumber']
        self._read()

    def _read(self):
        A = np.loadtxt(self._filename)
        self._points = A[:,0:3] # X Y Z
        self._values = A[:,3:] # remainder

    @property
    def points(self):
        return self._points

    @property
    def values(self):
        return self._values

    @property
    def value_names(self):
        return self._columns

    def points_generator(self):
        cnt = 0
        l = len(self.points)
        while cnt < l:
            yield self._points[cnt]

    def values_generator(self):
        cnt = 0
        l = len(self.values)
        while cnt < l:
            yield self._values[cnt]

if __name__=='__main__':
    r = DlpReader('../data/20170725_1_G.3dlp')
    p = r.points[0]
    print(type(p))
    print(p.shape)



import numpy as np
import pandas as pd

class RainStormReader(object):
    def __init__(self, filename):
        self._filename = filename
        self._values = None
        self._points = None
        self._columns = ['idx', 'frame_idx', 'x_coord', 'y_coord', 'I', 'sig_x',
       'sig_y', 'avg_brightness', 'res', 'res_Row', 'res_Col', 'roi_min',
       'Sum_signal', 'Sum_signal_ph', 'x_std', 'y_std',
       'ellip_xy']
        self._read()

    def _read(self):
        data = pd.read_csv(self._filename)
        pxtonm = max(data['x_nm']) / max(data['x_coord'])
        self._points = data[['x_nm', 'y_nm', 'z_coord']].values.copy()
        self._points[:,2] *= pxtonm
        self._values = data[['idx', 'frame_idx', 'x_coord', 'y_coord', 'I', 'sig_x',
       'sig_y', 'avg_brightness', 'res', 'res_Row', 'res_Col', 'roi_min',
       'Sum_signal', 'Sum_signal_ph', 'x_std', 'y_std',
       'ellip_xy']].values.copy()

    @property
    def points(self):
        return self._points

    @property
    def values(self):
        return self._values

    @property
    def value_names(self):
        return self._columns

    def points_generator(self):
        cnt = 0
        l = len(self.points)
        while cnt < l:
            yield self._points[cnt]

    def values_generator(self):
        cnt = 0
        l = len(self.values)
        while cnt < l:
            yield self._values[cnt]



import numpy as np
import pandas as pd

class AbbelightReader(object):
    def __init__(self, filename):
        self._filename = filename
        self._values = None
        self._points = None
        self._columns = ['id', 'frame', 'sigma1 [nm]',
       'sigma2 [nm]', 'intensity [photon]', 'offset [photon]',
       'bkgstd [photon]', 'uncertainty_xy [nm]', 'uncertainty_z [nm]',
       'ratio_SAF [photon]', 'z_SAF [nm]', 'z_Astigm [nm]']
        self._read()

    def _read(self):
        data = pd.read_csv(self._filename)
        self._points = data[['x [nm]', 'y [nm]', 'z [nm]']].values.copy()
        self._values = data[['id', 'frame','sigma1 [nm]',
       'sigma2 [nm]', 'intensity [photon]', 'offset [photon]',
       'bkgstd [photon]', 'uncertainty_xy [nm]', 'uncertainty_z [nm]',
       'ratio_SAF [photon]', 'z_SAF [nm]', 'z_Astigm [nm]']].values.copy()

    @property
    def points(self):
        return self._points

    @property
    def values(self):
        return self._values

    @property
    def value_names(self):
        return self._columns

    def points_generator(self):
        cnt = 0
        l = len(self.points)
        while cnt < l:
            yield self._points[cnt]

    def values_generator(self):
        cnt = 0
        l = len(self.values)
        while cnt < l:
            yield self._values[cnt]



from setuptools import setup
import versioneer

requirements = [
    'numpy', 'vtk', 'jupyter', 'pandas', 'requests', 'seaborn', 'versioneer'
]

setup(
    name='smlmvis',
    license='GPLv3',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Superresolution visualization of 3D protein localization data from a range of microscopes",
    author="Ben Cardoen",
    author_email='bcardoen@sfu.ca',
    url='https://github.com/bencardoen/smlmvis',
    download_url='https://github.com/bencardoen/smlmvis/archive/v0.0.5.tar.gz',
    packages=['smlmvis'],
    install_requires=requirements,
    keywords='smlmvis',
    classifiers=[
        'Programming Language :: Python :: 3.7',
    ]
)



from smlmvis.vtuwriter import VtuWriter
from collections import namedtuple
from numpy import sin as nsin
from numpy import cos as ncos
from numpy import arccos as narccos
from numpy import random as nrp
from numpy import sqrt
from numpy import pi
import numpy as np


class AbstractWriter:
    def __init__(self):
        pass

    @property
    def written(self):
        return self._written

    def write(self, name, points, values):
        VtuWriter(name, points, values)
        self._written = (points, values)


class Ellipsoid(AbstractWriter):
    def __init__(self, a=1, b=1, c=1, center=(0,0,0), perturbfactor = 1):
        self.a = a
        self.b = b
        self.c = c
        self.center = center
        self.perturbfactor = perturbfactor


    def _point(self, u, v):
        c = len(u)
        points = np.empty((c, 3))
        nv = nsin(v)
        points[:,0] = self.a * ncos(u) *nv + self.center[0]
        points[:,1] = self.b * nsin(u) *nv + self.center[1]
        points[:,2] = self.c * ncos(v) + self.center[2]#
        if self.perturbfactor != 1:
            points = points * (1 + np.random.random((c,3))) * self.perturbfactor
        return points

    def sample(self, rng=nrp, count=1, seed=0):
        rng.seed(seed)
        return self._point(rng.uniform(0, 1, count)*pi, rng.uniform(0, 1, count)*(2*pi))

    def write(self, count, name, values = None):
        points = self.sample(rng=nrp, count=count)
        if values is None:
            values = points
        super().write(name, points, values)

class HalfEllipsoid(AbstractWriter):
    def __init__(self, a, b, c, center, perturbfactor, zlimits, minima, maxima):
        '''
        Create an ellipsoid(a,b,c, center, perturbfactor) sliced by zlimits embedded in a plane specified by m, M.
        '''
        self.plane = Planar(minima, maxima)
        self.half_ellipse = OpenEllipsoid(a, b, c, center, perturbfactor, zlimits)

    def _in_ellipsoid(self, point):
        '''
        Test if point is contained within the member ellipsoid.
        #Fix non center intercept (i.e. also use Z intercept)
        '''
        x, y, _ = [(p - c)**2 for p,c in zip(point, self.half_ellipse.center)]
        rx, ry = self.half_ellipse.a, self.half_ellipse.b
        return (x / (rx**2)) + (y / (ry**2)) <= 1

    def write(self, count, name, values=None):
        ef_points = self.plane.sample(count=count)
        # Remove all points where x,y are within the area of the ellipsoid specified.
        # All x where d(x, center) < center[1]+-a)
        # All y where d(y, center) < b)
        mask = [~self._in_ellipsoid(ef_points[i,:]) for i in range(count)]
        e_points = ef_points[mask,:]
        masked = len(mask)
        p_points = self.half_ellipse.sample(rng=nrp, count=masked) # Not quite fair since sphere will be more depth

        points = np.concatenate((e_points, p_points))
        if values is None:
            values = points
        values = points
        super().write(name, points, values)


class OpenEllipsoid(Ellipsoid):
    def __init__(self, a=1, b=1, c=1, center=(0,0,0), perturbfactor =1, zrange=(-float('inf'), float('inf'))):
        assert(c>=1)
        super().__init__(a, b, c, center, perturbfactor)
        self.zrange = zrange

    @staticmethod
    def ztov(z, c, ctr):
        '''
        Given z coordinate, scale of Z radius and center point, return the angular threshold value correspoding with z (0-pi).
        # TODO fix c < 1 flipping range
        '''
        arg = (z-ctr)/c
        if arg >= 1:
            return 0
        if arg <= -1:
            return 0
        return narccos(arg)


    def sample(self, rng=nrp, count=1, seed=0):
        '''
        Sample count points (shape=count x 3) in 3D space using rng and seed.
        '''
        assert(count>0)
        rng.seed(seed)
        vm, vM = [OpenEllipsoid.ztov(z, self.c, self.center[2]) for z in self.zrange]
        h = int(count//2) if count > 1 else 1 # count pos int, int() floors
        fcount, scount = (h, h+1) if (count % 2) else (h, h)
        lowv = vM + rng.uniform(0, 1, fcount) * (vm - vM)      # explicit since rng api may not have range (random())
        highv = 2*pi - vm + rng.uniform(0, 1, scount) * (vm - vM)
        vpoints = np.concatenate((lowv, highv))
        upoints = rng.uniform(0,1,count)*pi
        return self._point(upoints, vpoints)


class GaussianEllipsoid(AbstractWriter):
    '''
    Create 3D Gaussian.
    '''
    def __init__(self, means, sigmas, seed = 0):
        self._rng = nrp
        nrp.seed(seed)
        self._means = means
        self._sigmas = sigmas

    def sample(self, count=1):
        points = np.empty((count, 3))
        for index, (mean, sigma) in enumerate(zip(self._means, self._sigmas)):
            points[:,index] = self._rng.normal(mean, sigma, count)
        return points

    def write(self, count, name):
        points = self.sample(count)
        values = np.arange(count).reshape(count, 1)
        super().write(name, points, values)

class Planar(AbstractWriter):
    def __init__(self, minima, maxima, seed=0):
        self._rng = nrp
        self._rng.seed(seed)
        self.minima = minima
        self.maxima = maxima

    def sample(self, count):
        points = np.empty((count, 3))
        for index, (min_ax, max_ax) in enumerate(zip(self.minima, self.maxima)):
            points[:,index] = self._rng.uniform(min_ax, max_ax, count)
        return points

    def write(self, count, name):
        points = self.sample(count)
        values = points
        super().write(name, points, values)


class Ellipsoids:
    def __init__(self, params):
        self.es = [Ellipsoid(*param) for param in params]

    def write(self, counts, names):
        k = len(self.es)
        for index, (ellipsoid, count, name) in enumerate(zip(self.es, counts, names)):
            values = np.full((count, 1), index)
            ellipsoid.write(count, name, values)

    def points(self):
        return [ell.written for ell in self.es]
